# Generated by Django 3.1.3 on 2020-12-03 21:07

import django.contrib.sites.managers
import django.db.models.manager
from django.db import migrations

import vendor.models.offer


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0016_auto_20201203_2011"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="offer",
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("on_site", django.contrib.sites.managers.CurrentSiteManager()),
                ("on_site_active", vendor.models.offer.ActiveCurrentSiteManager()),
            ],
        ),
    ]



import logging

from django.apps import apps
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.db import DatabaseError, IntegrityError, transaction
from django.db.models import Count, Q
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.translation import gettext as _
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, FormMixin, UpdateView
from django.views.generic.list import ListView

from vendor.config import VENDOR_PRODUCT_MODEL
from vendor.forms import (
    AddressForm,
    CreditCardForm,
    OfferForm,
    OfferSiteSelectForm,
    PriceFormSet,
    SiteSelectForm,
    StartDateForm,
    SubscriptionAddPaymentForm,
    SubscriptionForm,
)
from vendor.models import (
    CustomerProfile,
    Invoice,
    Offer,
    Payment,
    Receipt,
    Subscription,
)
from vendor.models.choice import InvoiceStatus, PaymentTypes, PurchaseStatus
from vendor.processors import (
    PRORATION_BEHAVIOUR_CHOICE,
    StripeProcessor,
    get_site_payment_processor,
)
from vendor.views.mixin import (
    PassRequestToFormKwargsMixin,
    SiteOnRequestFilterMixin,
    TableFilterMixin,
    get_site_from_request,
)

Product = apps.get_model(VENDOR_PRODUCT_MODEL)
logger = logging.getLogger(__name__)


#############
# Admin Views
class AdminDashboardView(LoginRequiredMixin, SiteOnRequestFilterMixin, ListView):
    """
    List of the most recent invoices generated on the current site.
    """

    template_name = "vendor/manage/dashboard.html"
    model = Invoice

    def get_queryset(self):
        """
        Return the most recent 10
        """
        queryset = super().get_queryset()

        return queryset[:10]


class AdminInvoiceListView(LoginRequiredMixin, SiteOnRequestFilterMixin, ListView):
    """
    List of all the invoices generated on the current site.
    """

    template_name = "vendor/manage/invoice_list.html"
    model = Invoice

    def get_queryset(self):
        """
        Ignores Cart state invoices
        """
        queryset = super().get_queryset()
        return queryset.order_by("updated")


class AdminInvoiceDetailView(LoginRequiredMixin, DetailView):
    """
    Details of an invoice generated on the current site.
    """

    template_name = "vendor/manage/invoice_detail.html"
    model = Invoice
    slug_field = "uuid"
    slug_url_kwarg = "uuid"


class AdminProductListView(
    LoginRequiredMixin, TableFilterMixin, SiteOnRequestFilterMixin, ListView
):
    """
    Creates a Product to be added to offers
    """

    template_name = "vendor/manage/products.html"
    model = Product
    paginate_by = 25

    def search_filter(self, queryset):
        return queryset.filter(name__icontains=self.request.GET.get("search_filter"))


class AdminProductUpdateView(LoginRequiredMixin, UpdateView):
    """
    Details of an invoice generated on the current site.
    """

    template_name = "vendor/manage/product.html"
    model = Product
    fields = ["sku", "name", "site", "available", "description", "meta"]
    slug_field = "uuid"
    slug_url_kwarg = "uuid"
    success_url = reverse_lazy("vendor_admin:manager-product-list")


class AdminProductCreateView(LoginRequiredMixin, CreateView):
    """
    Creates a Product to be added to offers
    """

    template_name = "vendor/manage/product.html"
    model = Product
    fields = ["sku", "name", "site", "available", "description", "meta"]
    success_url = reverse_lazy("vendor_admin:manager-product-list")

    def form_valid(self, form):
        new_product = form.save(commit=False)
        new_product.save()
        return redirect(self.success_url)


class AdminOfferListView(LoginRequiredMixin, SiteOnRequestFilterMixin, ListView):
    """
    Creates a Product to be added to offers
    """

    template_name = "vendor/manage/offers.html"
    model = Offer


class AdminOfferUpdateView(
    LoginRequiredMixin, PassRequestToFormKwargsMixin, UpdateView
):
    """
    Details of an invoice generated on the current site.
    """

    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm
    slug_field = "uuid"
    slug_url_kwarg = "uuid"
    template_name_suffix = "_update_form"

    def get_initial(self):
        return {"products": self.object.products.all()}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        offer_products = self.object.products.all()
        customers_who_own = CustomerProfile.objects.filter(
            receipts__products__in=offer_products
        )
        customers_who_dont_own = CustomerProfile.objects.all().exclude(
            pk__in=[customer_profile.pk for customer_profile in customers_who_own.all()]
        )

        context["customers_who_own"] = customers_who_own
        context["customers_who_dont_own"] = customers_who_dont_own

        context["formset"] = PriceFormSet(instance=self.object)

        return context

    def form_valid(self, form):
        price_formset = PriceFormSet(
            self.request.POST,
            self.request.FILES,
            instance=Offer.objects.get(uuid=self.kwargs["uuid"]),
        )

        offer = form.save(commit=False)

        if len(form.cleaned_data["products"]) > 1:
            offer.bundle = True
        offer.save()

        for product in form.cleaned_data["products"]:
            offer.products.add(product)

        if price_formset.has_changed() and not price_formset.is_valid():
            return render(
                self.request,
                self.template_name,
                {"form": form, "formset": price_formset},
            )
        elif price_formset.is_valid():
            for price_form in price_formset:
                price = price_form.save(commit=False)
                price.offer = offer
                if price_form.cleaned_data["price_select"] == "free":
                    price.cost = 0
                price.save()
        else:
            return render(
                self.request,
                self.template_name,
                {"form": form, "formset": price_formset},
            )

        return redirect("vendor_admin:manager-offer-list")


class AdminOfferCreateView(
    LoginRequiredMixin, PassRequestToFormKwargsMixin, CreateView
):
    """
    Creates a Product to be added to offers
    """

    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["formset"] = PriceFormSet()

        return context

    def post(self, request):
        offer_form = self.form_class(request.POST, request=request)
        price_formset = PriceFormSet(request.POST)

        if not offer_form.is_valid():
            return render(
                request,
                self.template_name,
                {"form": offer_form, "formset": price_formset},
            )

        offer = offer_form.save(commit=False)
        if len(offer_form.cleaned_data["products"]) > 1:
            offer.bundle = True

        offer.save()
        for product in offer_form.cleaned_data["products"]:
            offer.products.add(product)

        if price_formset.has_changed() and not price_formset.is_valid():
            return render(
                request,
                self.template_name,
                {"form": offer_form, "formset": price_formset},
            )
        elif price_formset.has_changed() and price_formset.is_valid():
            product_currencies = {}
            # price_currency = [price_form.cleaned_data['currency']
            #                   for price_form in price_formset]

            for product in Product.objects.filter(
                pk__in=offer_form.cleaned_data["products"]
            ):
                for currency in product.meta["msrp"].keys():
                    product_currencies[currency] = currency

            for price_form in price_formset:
                if price_form.cleaned_data["currency"] not in product_currencies:
                    price_formset[0].add_error("currency", _("Invalid currency"))
                    return render(
                        request,
                        self.template_name,
                        {"form": offer_form, "formset": price_formset},
                    )

            for price_form in price_formset:
                price = price_form.save(commit=False)
                if price_form.cleaned_data["price_select"] == "free":
                    price.cost = 0
                price.offer = offer
                price.save()

        return redirect("vendor_admin:manager-offer-list")


class AdminSubscriptionListView(LoginRequiredMixin, ListView):
    """
    List of all the invoices generated on the current site.
    """

    template_name = "vendor/manage/receipt_list.html"
    model = Subscription

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset.filter(profile__site=get_site_from_request(self.request))


class AdminStripeSubscriptionListView(LoginRequiredMixin, ListView):
    """
    Lists Stripe Subscriptions from the requests, site.
    """

    template_name = "vendor/manage/receipt_list.html"
    model = Subscription

    def get_queryset(self):
        site = get_site_from_request(self.request)
        qs = Subscription.objects.filter(
            profile__site=site, gateway_id__startswith="sub_"
        ).order_by("status", "-pk")
        return qs


class AdminStripeSubscriptionReCreate(LoginRequiredMixin, FormMixin, DetailView):
    """
    View will get the Stripe Subscription Detail and on post
    will cancel the existing Subscription and create a new one
    to update any application fees or other price or offer
    related details
    """

    template_name = "vendor/manage/stripe_subscription_detail.html"
    model = Subscription
    slug_field = "uuid"
    slug_url_kwarg = "uuid"
    form_class = StartDateForm
    success_url = reverse_lazy("vendor_admin:manager-stripe-subscriptions")

    def post(self, request, **kwargs):
        form = StartDateForm(request.POST)

        if not form.is_valid():
            context = self.get_context_data(**kwargs)
            context["form"] = form
            return render(request, self.template_name, context)

        site = get_site_from_request(request)
        subscription = self.get_object()
        subscription_extras = {
            "billing_cycle_anchor": form.cleaned_data["start_date"],
            "proration_behavior": PRORATION_BEHAVIOUR_CHOICE.NONE.value,
        }
        invoice = subscription.profile.get_cart()
        invoice.empty_cart()
        invoice.add_offer(subscription.get_offer())

        stripe = StripeProcessor(site, invoice)

        stripe_payment_methods = stripe.get_customer_payment_methods(
            subscription.profile.meta["stripe_id"]
        )
        if not stripe_payment_methods:
            raise ObjectDoesNotExist(
                f"stripe_payment_method does not exist for stripe_customer: {subscription.profile.meta['stripe_id']}"
            )

        stripe.create_subscription(stripe_payment_methods[0].id, subscription_extras)

        if stripe.transaction_succeeded:
            stripe.subscription_cancel(subscription)

        invoice.empty_cart()

        return redirect(
            reverse(
                "vendor_admin:manager-subscription",
                kwargs={"uuid": str(self.get_object().uuid)},
            )
        )


class AdminSubscriptionDetailView(LoginRequiredMixin, DetailView):
    """
    Gets all Customer Profile information for quick lookup and management
    """

    template_name = "vendor/manage/subscription_detail.html"
    model = Subscription
    slug_field = "uuid"
    slug_url_kwarg = "uuid"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        subscription = context["object"]
        payment = subscription.payments.first()

        context["payment"] = payment
        context["payment_form"] = CreditCardForm(
            initial={"payment_type": PaymentTypes.CREDIT_CARD}
        )
        if payment and payment.billing_address:
            context["billing_form"] = AddressForm(instance=payment.billing_address)

        context["payments"] = subscription.payments.order_by("-invoice__pk")
        context["receipts"] = subscription.receipts.order_by("-order_item__invoice__pk")

        return context


class AdminSubscriptionCreateView(LoginRequiredMixin, TemplateView):
    """
    Gets all Customer Profile information for quick lookup and management
    """

    template_name = "vendor/manage/subscription_create.html"
    success_url = reverse_lazy("vendor_admin:manager-subscription-create")

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        if "site" in request.GET:
            site_form = SiteSelectForm(request.GET)

            if not site_form.is_valid():
                context["site_form"] = site_form
                return render(request, self.template_name, context)

            context["subscription_form"] = SubscriptionForm(
                initial={"site": site_form.cleaned_data["site"]}
            )
            context["site_form"] = site_form

            return render(request, self.template_name, context)

        context["site_form"] = SiteSelectForm()

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        subscription_form = SubscriptionForm(request.POST)

        if not subscription_form.is_valid():
            context["subscription_form"] = subscription_form
            return render(request, self.template_name, context)

        subscription = Subscription.objects.create(  # noqa: F841
            profile=subscription_form.cleaned_data["profile"],
            gateway_id=subscription_form.cleaned_data["subscription_id"],
            status=subscription_form.cleaned_data["status"],
        )
        messages.info(request, _("Subscription Created"))

        return redirect(self.success_url)


class AdminSubscriptionAddPaymentView(LoginRequiredMixin, TemplateView):
    template_name = "vendor/manage/subscription_add_payment.html"
    success_url = reverse_lazy("vendor_admin:manager-subscriptions")

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        subscription = Subscription.objects.get(uuid=kwargs.get("uuid_subscription"))
        profile = CustomerProfile.objects.get(uuid=kwargs.get("uuid_profile"))

        if "site" in request.GET:
            offer_site_form = OfferSiteSelectForm(request.GET)

            if not offer_site_form.is_valid():
                context["offer_site_form"] = offer_site_form
                return render(request, self.template_name, context)

            context["form"] = SubscriptionAddPaymentForm(
                initial={
                    "offer": offer_site_form.cleaned_data["offer"],
                    "subscription": subscription,
                    "profile": profile,
                }
            )

            context["offer_site_form"] = offer_site_form

            return render(request, self.template_name, context)

        context["offer_site_form"] = OfferSiteSelectForm()

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        offer_site_form = OfferSiteSelectForm(request.GET)

        if not offer_site_form.is_valid():
            context["offer_site_form"] = (
                payment_form  # TODO: fix this.  It's not defined or imported.  # noqa: F821,E501
            )
            return render(request, self.template_name, context)

        payment_form = SubscriptionAddPaymentForm(
            request.POST, site=offer_site_form.cleaned_data["site"]
        )

        if not payment_form.is_valid():
            context["form"] = payment_form
            return render(request, self.template_name, context)

        payment = payment_form.save(commit=False)
        offer = payment_form.cleaned_data["offer"]

        invoice = payment.profile.get_cart_or_checkout_cart()
        invoice.empty_cart()
        invoice.status = InvoiceStatus.COMPLETE
        invoice.add_offer(offer)

        try:
            with transaction.atomic():
                payment.invoice = invoice
                payment.amount = invoice.total
                payment.save()

                if payment.success and payment.status == PurchaseStatus.SETTLED:
                    receipt = Receipt.objects.create(
                        transaction=payment.transaction,
                        order_item=invoice.order_items.first(),
                        profile=payment.profile,
                        start_date=offer.get_offer_start_date(payment.submitted_date),
                        end_date=offer.get_offer_end_date(payment.submitted_date),
                        subscription=payment.subscription,
                    )
                    receipt.products.add(offer.products.first())

            messages.info(request, _("Payment Added to Subscription"))

        except (IntegrityError, DatabaseError, Exception) as exce:
            logger.error(f"AdminSubscriptionCreateView error: {exce}")
            messages.error(request, "failed to add payment to subscription")
            return render(request, self.template_name, context)

        return redirect(request.META.get("HTTP_REFERER", self.success_url))


class AdminProfileListView(
    LoginRequiredMixin, TableFilterMixin, SiteOnRequestFilterMixin, ListView
):
    """
    List of CustomerProfiles on site
    """

    template_name = "vendor/manage/profile_list.html"
    model = CustomerProfile
    paginate_by = 100

    def search_filter(self, queryset):
        search_value = self.request.GET.get("search_filter")
        return queryset.filter(
            Q(pk__icontains=search_value)
            | Q(user__email__icontains=search_value)
            | Q(user__username__icontains=search_value)
        )

    def get_paginated_by(self, queryset):
        if "paginate_by" in self.request.kwargs:
            return self.kwargs["paginate_by"]
        return self.paginate_by

    def get_queryset(self):
        queryset = super().get_queryset()

        return queryset.order_by("pk")


class AdminProfileDetailView(LoginRequiredMixin, DetailView):
    """
    Gets all Customer Profile information for quick lookup and management
    """

    template_name = "vendor/manage/profile_detail.html"
    model = CustomerProfile
    slug_field = "uuid"
    slug_url_kwarg = "uuid"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["free_offers"] = Offer.objects.filter(
            prices__cost=0, site=self.object.site
        )
        context["invoices"] = self.object.invoices.order_by("-created")

        return context


class AdminManualSubscriptionRenewal(LoginRequiredMixin, DetailView):
    success_url = reverse_lazy("vendor_admin:manage-profiles")
    model = Subscription
    slug_field = "uuid"
    slug_url_kwarg = "uuid"

    def post(self, request, *args, **kwargs):
        subscription = Subscription.objects.get(uuid=self.kwargs["uuid"])
        submitted_datetime = timezone.now()
        site = get_site_from_request(request)
        customer_profile = request.user.customer_profile

        invoice = Invoice.objects.create(
            profile=customer_profile,
            site=site,
            ordered_date=submitted_datetime,
            status=InvoiceStatus.COMPLETE,
        )
        invoice.add_offer(  # TODO: fix this.  It's not defined or imported.  # noqa: F821,E501
            offer  # TODO: fix this.  It's not defined or imported.  # noqa: F821,E501
        )
        invoice.save()

        transaction_id = timezone.now().strftime("%Y-%m-%d_%H-%M-%S-Manual-Renewal")

        processor = get_site_payment_processor(invoice.site)(invoice.site, invoice)
        processor.renew_subscription(
            subscription, transaction_id, PurchaseStatus.CAPTURED, submitted_datetime
        )

        messages.info(request, _("Subscription Renewed"))
        return redirect(request.META.get("HTTP_REFERER", self.success_url))


class PaymentWithNoReceiptListView(LoginRequiredMixin, ListView):
    template_name = "vendor/manage/payment_list.html"
    model = Payment

    def get_queryset(self):
        site = get_site_from_request(self.request)
        return [
            payment
            for payment in Payment.objects.filter(invoice__site=site, success=True)
            if payment.get_receipt() is None
        ]

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = _("Payments with no Receipts")
        return context


class PaymentWithNoOrderItemsListView(LoginRequiredMixin, ListView):
    template_name = "vendor/manage/payment_list.html"
    model = Payment

    def get_queryset(self):
        site = get_site_from_request(self.request)
        return (
            Payment.objects.filter(invoice__site=site, success=True)
            .annotate(order_item_count=Count("invoice__order_items"))
            .filter(order_item_count=0)
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = _("Payments with no Order Items")
        return context



#  Copyright (c) 2018-present Justin Mayer
#  Copyright (c) 2008—2016 Andy Mikhailenko
#
#  This file is part of django-autoslug.
#
#  django-autoslug is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README for copying conditions.
#

# django
from django.conf import settings
from django.db.models.fields import SlugField
from django.db.models.signals import post_save

# 3rd-party
try:
    from modeltranslation import utils as modeltranslation_utils
except ImportError:
    modeltranslation_utils = None

from vendor import utils

# this app
from vendor.settings import autoslug_modeltranslation_enable, slugify

__all__ = ["AutoSlugField"]

SLUG_INDEX_SEPARATOR = "-"  # the "-" in "foo-2"


class AutoSlugField(SlugField):
    """
    AutoSlugField is an extended SlugField able to automatically resolve name
    clashes.

    AutoSlugField can also perform the following tasks on save:

    - populate itself from another field (using `populate_from`),
    - use custom `slugify` function (using `slugify` or :doc:`settings`), and
    - preserve uniqueness of the value (using `unique` or `unique_with`).

    None of the tasks is mandatory, i.e. you can have auto-populated non-unique
    fields, manually entered unique ones (absolutely unique or within a given
    date) or both.

    Uniqueness is preserved by checking if the slug is unique with given constraints
    (`unique_with`) or globally (`unique`) and adding a number to the slug to make
    it unique.

    :param always_update: boolean: if True, the slug is updated each time the
        model instance is saved. Use with care because `cool URIs don't
        change`_ (and the slug is usually a part of object's URI). Note that
        even if the field is editable, any manual changes will be lost when
        this option is activated.
    :param populate_from: string or callable: if string is given, it is considered
        as the name of attribute from which to fill the slug. If callable is given,
        it should accept `instance` parameter and return a value to fill the slug
        with.
    :param sep: string: if defined, overrides default separator for automatically
        incremented slug index (i.e. the "-" in "foo-2").
    :param slugify: callable: if defined, overrides `AUTOSLUG_SLUGIFY_FUNCTION`
        defined in :doc:`settings`.
    :param unique: boolean: ensure total slug uniqueness (unless more precise
        `unique_with` is defined).
    :param unique_with: string or tuple of strings: name or names of attributes
        to check for "partial uniqueness", i.e. there will not be two objects
        with identical slugs if these objects share the same values of given
        attributes. For instance, ``unique_with='pub_date'`` tells AutoSlugField
        to enforce slug uniqueness of all items published on given date. The
        slug, however, may reappear on another date. If more than one field is
        given, e.g. ``unique_with=('pub_date', 'author')``, then the same slug may
        reappear within a day or within some author's articles but never within
        a day for the same author. Foreign keys are also supported, i.e. not only
        `unique_with='author'` will do, but also `unique_with='author__name'`.

    .. _cool URIs don't change: http://w3.org/Provider/Style/URI.html

    .. note:: always place any slug attribute *after* attributes referenced
        by it (i.e. those from which you wish to `populate_from` or check
        `unique_with`). The reasoning is that autosaved dates and other such
        fields must be already processed before using them in the AutoSlugField.

    Example usage:

    .. code-block:: python

        from django.db import models
        from vendor.fields import AutoSlugField

        class Article(models.Model):
            '''An article with title, date and slug. The slug is not totally
            unique but there will be no two articles with the same slug within
            any month.
            '''
            title = models.CharField(max_length=200)
            pub_date = models.DateField(auto_now_add=True)
            slug = AutoSlugField(populate_from='title', unique_with='pub_date__month')


    More options:

    .. code-block:: python

        # slugify but allow non-unique slugs
        slug = AutoSlugField()

        # globally unique, silently fix on conflict ("foo" --> "foo-1".."foo-n")
        slug = AutoSlugField(unique=True)

        # autoslugify value from attribute named "title"; editable defaults to False
        slug = AutoSlugField(populate_from='title')

        # same as above but force editable=True
        slug = AutoSlugField(populate_from='title', editable=True)

        # ensure that slug is unique with given date (not globally)
        slug = AutoSlugField(unique_with='pub_date')

        # ensure that slug is unique with given date AND category
        slug = AutoSlugField(unique_with=('pub_date','category'))

        # ensure that slug is unique with an external object
        # assuming that author=ForeignKey(Author)
        slug = AutoSlugField(unique_with='author')

        # ensure that slug is unique with a subset of external objects (by lookups)
        # assuming that author=ForeignKey(Author)
        slug = AutoSlugField(unique_with='author__name')

        # mix above-mentioned behaviour bits
        slug = AutoSlugField(populate_from='title', unique_with='pub_date')

        # minimum date granularity is shifted from day to month
        slug = AutoSlugField(populate_from='title', unique_with='pub_date__month')

        # autoslugify value from a dynamic attribute (i.e. a method)
        slug = AutoSlugField(populate_from='get_full_name')

        # autoslugify value from a custom callable
        # (ex. usage: user profile models)
        slug = AutoSlugField(populate_from=lambda instance: instance.user.get_full_name())

        # specify model manager for looking up slugs shared by subclasses

        class Article(models.Model):
            '''An article with title, date and slug. The slug is not totally
            unique but there will be no two articles with the same slug within
            any month.
            '''
            objects = models.Manager()
            title = models.CharField(max_length=200)
            slug = AutoSlugField(populate_from='title', unique_with='pub_date__month', manager=objects)

        class NewsArticle(Article):
            pass

        # autoslugify value using custom `slugify` function
        from vendor.settings import slugify as default_slugify
        def custom_slugify(value):
            return default_slugify(value).replace('-', '_')
        slug = AutoSlugField(slugify=custom_slugify)

    """

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = kwargs.get("max_length", 50)

        # autopopulated slug is not editable unless told so
        self.populate_from = kwargs.pop("populate_from", None)
        if self.populate_from:
            kwargs.setdefault("editable", False)

        # unique_with value can be string or tuple
        self.unique_with = kwargs.pop("unique_with", ())
        if isinstance(self.unique_with, str):
            self.unique_with = (self.unique_with,)

        self.slugify = kwargs.pop("slugify", slugify)
        assert hasattr(self.slugify, "__call__")

        self.index_sep = kwargs.pop("sep", SLUG_INDEX_SEPARATOR)

        if self.unique_with:
            # we will do "manual" granular check below
            kwargs["unique"] = False

        # Set db_index=True unless it's been set manually.
        if "db_index" not in kwargs:
            kwargs["db_index"] = True

        # A boolean instructing the field to accept Unicode letters in
        # addition to ASCII letters. Defaults to False.
        self.allow_unicode = kwargs.pop("allow_unicode", False)

        # When using model inheritance, set manager to search for matching
        # slug values
        self.manager = kwargs.pop("manager", None)
        self.manager_name = kwargs.pop("manager_name", None)

        self.always_update = kwargs.pop("always_update", False)
        super(SlugField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        if self.max_length == 50:
            kwargs.pop("max_length", None)

        if self.populate_from is not None:
            kwargs["populate_from"] = self.populate_from
            if self.editable is not False:
                kwargs["editable"] = self.editable

        if self.unique_with != ():
            kwargs["unique_with"] = self.unique_with
            kwargs.pop("unique", None)

        if self.slugify != slugify:
            kwargs["slugify"] = self.slugify

        if self.index_sep != SLUG_INDEX_SEPARATOR:
            kwargs["sep"] = self.index_sep

        kwargs.pop("db_index", None)

        if self.manager is not None:
            kwargs["manager"] = self.manager

        if self.manager_name is not None:
            kwargs["manager_name"] = self.manager_name

        if self.always_update:
            kwargs["always_update"] = self.always_update

        if "manager" in kwargs:
            del kwargs["manager"]

        return name, path, args, kwargs

    def pre_save(self, instance, add):

        # get currently entered slug
        value = self.value_from_object(instance)

        if self.manager is not None:
            manager = self.manager
        elif self.manager_name is not None:
            manager = getattr(self.model, self.manager_name)
        else:
            manager = None

        # autopopulate
        if self.always_update or (self.populate_from and not value):
            value = utils.get_prepopulated_value(self, instance)

            # pragma: nocover
            if __debug__ and not value and not self.blank:
                print(
                    "Failed to populate slug %s.%s from %s"
                    % (instance._meta.object_name, self.name, self.populate_from)
                )

        slug = None
        if value:
            slug = self.slugify(value)
        if not slug:
            slug = None

            if not self.blank:
                slug = instance._meta.model_name
            elif not self.null:
                slug = ""

        if slug:
            slug = self.slugify(utils.crop_slug(self, slug))

            # ensure the slug is unique (if required)
            if self.unique or self.unique_with:
                slug = utils.generate_unique_slug(self, instance, slug, manager)

            assert slug, "value is filled before saving"

        # make the updated slug available as instance attribute
        setattr(instance, self.name, slug)

        # modeltranslation support
        if (
            "modeltranslation" in settings.INSTALLED_APPS
            and not hasattr(self.populate_from, "__call__")
            and autoslug_modeltranslation_enable
        ):
            post_save.connect(modeltranslation_update_slugs, sender=type(instance))

        return slug


def modeltranslation_update_slugs(sender, **kwargs):
    # https://bitbucket.org/neithere/django-autoslug/pull-request/11/modeltranslation-support-fix-issue-19/
    # http://django-modeltranslation.readthedocs.org
    #
    # TODO: tests
    #
    if not modeltranslation_utils:
        return

    instance = kwargs["instance"]
    slugs = {}

    for field in instance._meta.fields:
        if type(field) is not AutoSlugField:
            continue
        if not field.populate_from:
            continue
        for lang in settings.LANGUAGES:
            lang_code, _ = lang
            lang_code = lang_code.replace("-", "_")

            populate_from_localized = modeltranslation_utils.build_localized_fieldname(
                field.populate_from, lang_code
            )
            field_name_localized = modeltranslation_utils.build_localized_fieldname(
                field.name, lang_code
            )

            # The source field or the slug field itself may not be registered
            # with translator
            if not hasattr(instance, populate_from_localized):
                continue
            if not hasattr(instance, field_name_localized):
                continue

            populate_from_value = getattr(instance, populate_from_localized)
            field_value = getattr(instance, field_name_localized)

            if not field_value or field.always_update:
                slug = field.slugify(populate_from_value)
                slugs[field_name_localized] = slug

    sender.objects.filter(pk=instance.pk).update(**slugs)



# Generated by Django 3.2.16 on 2023-01-18 18:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0040_auto_20221020_1617"),
    ]

    operations = [
        migrations.AddField(
            model_name="offer",
            name="billing_start_date",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="The start date to begin billing",
                null=True,
                verbose_name="Billing Start Date",
            ),
        ),
    ]



# App Settings
from django import forms
from django.conf import settings
from django.contrib.sites.models import Site
from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _
from siteconfigs.config import SiteConfigBaseClass


class SiteSelectForm(forms.Form):
    site = forms.ModelChoiceField(queryset=Site.objects.all())


class SupportedPaymentProcessor(TextChoices):
    PROMO_CODE_BASE = ("base.PaymentProcessorBase", _("Default Processor"))
    AUTHORIZE_NET = ("authorizenet.AuthorizeNetProcessor", _("Authorize.Net"))
    STRIPE = ("stripe.StripeProcessor", _("Stripe"))


class PaymentProcessorForm(SiteSelectForm):
    payment_processor = forms.CharField(
        label=_("Payment Processor"),
        widget=forms.Select(choices=SupportedPaymentProcessor.choices),
    )


class PaymentProcessorSiteConfig(SiteConfigBaseClass):
    label = "Payment Processor"
    default = {"payment_processor": "base.PaymentProcessorBase"}
    form_class = PaymentProcessorForm
    key = ""
    instance = None

    def __init__(self, site=None):
        if site is None:
            site = Site.objects.get_current()
        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site)
        if not self.instance:
            self.default["payment_processor"] = VENDOR_PAYMENT_PROCESSOR

    def get_form(self):
        return self.form_class(initial=self.get_initials())

    def get_initials(self):
        initial = super().get_initials()
        initial["site"] = (self.site.pk, self.site.domain)
        if self.instance:
            initial["payment_processor"] = [
                choice
                for choice in SupportedPaymentProcessor.choices
                if choice[0] == self.instance.value["payment_processor"]
            ][0]
        else:
            initial["payment_processor"] = SupportedPaymentProcessor.choices[0]
        return initial

    def get_selected_processor(self):
        if self.instance:
            return [
                choice
                for choice in SupportedPaymentProcessor.choices
                if choice[0] == self.instance.value["payment_processor"]
            ][0]
        return SupportedPaymentProcessor.choices[0]  # Return Default Processors

    def get_key_value(self, key):
        """
        Return the value for a given key from the config instance or default.
        """
        if self.instance and hasattr(self.instance, "value"):
            return self.instance.value.get(key)
        return self.default.get(key)

    def save(self, value=None, key=None):
        """
        Save a single key-value pair to the config instance, or all if no key provided.
        """
        current_value = getattr(self, "value", None)
        if value is not None and key is not None:
            data = current_value.copy() if current_value else self.default.copy()
            data[key] = value
            self.value = data
            super().save()
        else:
            super().save()


class StripeConnectAccountForm(SiteSelectForm):
    account_number = forms.CharField(max_length=120)


class StripeConnectAccountConfig(SiteConfigBaseClass):
    label = "Stripe Connect Account"
    default = {"stripe_connect_account": None}
    form_class = StripeConnectAccountForm
    key = ""
    instance = None

    def __init__(self, site=None):
        if site is None:
            site = Site.objects.get_current()
        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site)

    def save(self, value=None, key=None):
        """
        Save a single key-value pair to the config instance, or all if no key provided.
        """
        current_value = getattr(self, "value", None)
        if value is not None and key is not None:
            data = current_value.copy() if current_value else self.default.copy()
            data[key] = value
            self.value = data
            super().save()
        else:
            super().save()


class VendorSiteCommissionForm(SiteSelectForm):
    commission = forms.IntegerField(min_value=0, max_value=100)


class VendorSiteCommissionConfig(SiteConfigBaseClass):
    label = "Vendor Site Commission"
    default = {"commission": None}
    form_class = VendorSiteCommissionForm
    key = ""
    instance = None

    def __init__(self, site=None):
        if site is None:
            site = Site.objects.get_current()
        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site)

    def save(self, value=None, key=None):
        """
        Save a single key-value pair to the config instance, or all if no key provided.
        """
        current_value = getattr(self, "value", None)
        if value is not None and key is not None:
            data = current_value.copy() if current_value else self.default.copy()
            data[key] = value
            self.value = data
            super().save()
        else:
            super().save()


VENDOR_PRODUCT_MODEL = getattr(settings, "VENDOR_PRODUCT_MODEL", "vendor.Product")

VENDOR_PAYMENT_PROCESSOR = getattr(
    settings, "VENDOR_PAYMENT_PROCESSOR", "dummy.DummyProcessor"
)

DEFAULT_CURRENCY = getattr(settings, "DEFAULT_CURRENCY", "usd")

AVAILABLE_CURRENCIES = getattr(
    settings, "AVAILABLE_CURRENCIES", {"usd": _("USD Dollars")}
)

VENDOR_STATE = getattr(settings, "VENDOR_STATE", "DEBUG")

# Encryption settings
VENDOR_DATA_ENCODER = getattr(
    settings, "VENDOR_DATA_ENCODER", "vendor.encrypt.cleartext"
)

ENABLE_STRIPE_SIGNALS = getattr(settings, "ENABLE_STRIPE_SIGNALS", False)

STRIPE_BASE_COMMISSION = getattr(
    settings, "STRIPE_BASE_COMMISSION", {"percentage": 2.9, "fixed": 0.3}
)

STRIPE_RECURRING_COMMISSION = getattr(
    settings, "STRIPE_RECURRING_COMMISSION", {"percentage": 0.5}
)



import uuid

from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from vendor.models.base import CreateUpdateModelBase, SoftDeleteModelBase


class Receipt(SoftDeleteModelBase, CreateUpdateModelBase):
    """
    A link for all the purchases a user has made. Contains subscription start and end date.
    This is generated for each item a user purchases so it can be checked in other code.
    """

    uuid = models.UUIDField(
        _("UUID"),
        editable=False,
        unique=True,
        default=uuid.uuid4,
        null=False,
        blank=False,
    )
    profile = models.ForeignKey(
        "vendor.CustomerProfile",
        verbose_name=_("Purchase Profile"),
        on_delete=models.CASCADE,
        related_name="receipts",
    )
    order_item = models.ForeignKey(
        "vendor.OrderItem",
        verbose_name=_("Order Item"),
        on_delete=models.CASCADE,
        related_name="receipts",
    )
    start_date = models.DateTimeField(_("Start Date"), blank=True, null=True)
    end_date = models.DateTimeField(_("End Date"), blank=True, null=True)
    vendor_notes = models.JSONField(
        _("Vendor Notes"), default=dict, blank=True, null=True
    )
    transaction = models.CharField(
        _("Transaction"), max_length=80, blank=True, null=True
    )
    subscription = models.ForeignKey(
        "vendor.Subscription",
        verbose_name=_("Subscription"),
        on_delete=models.CASCADE,
        related_name="receipts",
        blank=True,
        null=True,
        default=None,
    )
    meta = models.JSONField(_("Meta"), default=dict, blank=True, null=True)

    class Meta:
        verbose_name = "Receipt"
        verbose_name_plural = "Receipts"

    def __str__(self):
        return f"{self.profile.user.username} - {self.order_item.offer.name}"

    def get_absolute_url(self):
        return reverse("vendor:customer-receipt", kwargs={"uuid": self.uuid})

    def is_on_trial(self, today=timezone.now()):

        if "trial" in self.transaction and self.end_date >= today:
            return True

        return False



# Generated by Django 3.1.3 on 2021-09-14 12:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0025_auto_20210914_0025"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="deleted",
            field=models.BooleanField(default=False, verbose_name="Deleted"),
        ),
        migrations.AddField(
            model_name="payment",
            name="deleted",
            field=models.BooleanField(default=False, verbose_name="Deleted"),
        ),
        migrations.AddField(
            model_name="receipt",
            name="deleted",
            field=models.BooleanField(default=False, verbose_name="Deleted"),
        ),
    ]



import csv
from itertools import chain

from django.contrib import messages
from django.contrib.sites.models import Site
from django.http import StreamingHttpResponse
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import FormMixin
from django.views.generic.list import BaseListView

from vendor.forms import DateRangeForm
from vendor.models import Invoice, Receipt
from vendor.utils import get_site_from_request


class Echo:
    """An object that implements just the write method of the file-like
    interface.
    """

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


class CSVStreamRowView(BaseListView):
    """A base view for displaying a list of objects."""

    filename = "receipt_list.csv"

    def get_queryset(self):
        return super().get_queryset()

    def get_row_data(self):
        header = [
            ["ROW_NAME", "ROW_COUNT"]
        ]  # Has to be a list inside an iterable (another list) for the chain to work.
        rows = (
            ["Row {}".format(idx), str(idx)] for idx in range(500)
        )  # Dummy data to show that its working.
        return chain(header, rows)

    def get(self, request, *args, **kwargs):
        rows = self.get_row_data()
        pseudo_buffer = Echo()
        writer = csv.writer(pseudo_buffer)
        # TODO: Figure out what to prepend the header without breaking the streaming generator
        response = StreamingHttpResponse(
            (writer.writerow(row) for row in rows), content_type="text/csv"
        )

        # Set the filename
        response["Content-Disposition"] = 'attachment; filename="{}"'.format(
            self.filename
        )
        return response


class ReceiptListCSV(FormMixin, CSVStreamRowView):
    filename = "receipts.csv"
    model = Receipt
    form_class = DateRangeForm
    header = [
        [
            _("Order ID"),
            _("Title"),
            _("Order Date"),
            _("Order Status"),
            _("Order Total"),
            _("Product ID"),
            _("Product Name"),
            _("Quantity"),
            _("Item Cost"),
            _("Item Total"),
            _("Discount Amount "),
            _("Coupon Code"),
            _("Coupons Used"),
            _("Total Discount Amount"),
            _("Refund ID"),
            _("Refund Total"),
            _("Refund Amounts"),
            _("Refund Reason"),
            _("Refund Date"),
            _("Refund Author Email"),
            _("Date Type"),
            _("Dates"),
        ]
    ]

    def get_queryset(self):
        form = self.form_class(data=self.request.POST)
        start_date = form.data.get("start_date", None)
        end_date = form.data.get("end_date", None)
        if start_date and end_date:
            return self.model.objects.filter(
                profile__site=Site.objects.get_current(),
                order_item__invoice__created__gte=start_date,
                order_item__invoice__created__lte=end_date,
            )
        elif start_date and not end_date:
            return self.model.objects.filter(
                profile__site=Site.objects.get_current(),
                order_item__invoice__created__gte=start_date,
            )
        else:
            return self.model.objects.filter(
                profile__site=Site.objects.get_current()
            )  # Return receipts only for profiles on this site

    def get_row_data(self):
        object_list = self.get_queryset()
        rows = [
            [
                str(obj.order_item.invoice.pk),  # Order ID
                obj.order_item.invoice,  # Title
                obj.order_item.invoice.created.isoformat(),  # Oder Date
                obj.get_status_display(),  # Order Status
                obj.order_item.invoice.total,  # Order Total
                str(obj.products.first().id),  # Product ID
                obj.products.first().name,  # Product Name
                obj.order_item.quantity,  # Quantity
                obj.order_item.price,  # Item Cost
                obj.order_item.total,  # Item Total
                "",  # TODO: Discount Amount
                "",  # TODO: Coupon Code
                "",  # TODO: Coupons Used
                "",  # TODO: Total Discount Amount
                "",  # TODO: Refund ID
                "",  # TODO: Refund Total
                "",  # TODO: Refound Amounts
                "",  # TODO: Refund Reason
                "",  # TODO: Refund Date
                "",  # TODO: Refund Author Email
                "multiple" if obj.end_date is None else "range",  # TODO: Date Type
                " ".join(
                    [
                        "" if obj.start_date is None else f"{obj.start_date:%Y-%m-%d}",
                        "" if obj.end_date is None else f"{obj.end_date:%Y-%m-%d}",
                    ]
                ),
            ]
            for obj in object_list
        ]
        return chain(self.header, rows)

    def post(self, request, *args, **kwargs):
        form = self.form_class(request.POST)

        if not form.is_valid():
            messages.info(self.request, ",".join([error for error in form.errors]))
            return redirect(request.META.get("HTTP_REFERER", self.success_url))

        rows = self.get_row_data()
        pseudo_buffer = Echo()
        writer = csv.writer(pseudo_buffer)
        # TODO: Figure out what to prepend the header without breaking the streaming generator
        response = StreamingHttpResponse(
            (writer.writerow(row) for row in rows), content_type="text/csv"
        )

        # Set the filename
        response["Content-Disposition"] = 'attachment; filename="{}"'.format(
            self.filename
        )
        return response


class InvoiceListCSV(CSVStreamRowView):
    filename = "invoices.csv"
    model = Invoice

    def get_queryset(self):
        # TODO: Update to handle ranges from a POST
        if hasattr(self.request, "site"):
            return self.model.objects.filter(site=get_site_from_request(self.request))
        return self.model.on_site.all()

    def get_row_data(self):
        object_list = self.get_queryset()
        header = [
            ["INVOICE_ID", "CREATED_TIME(ISO)", "USERNAME", "CURRENCY", "TOTAL"]
        ]  # Has to be a list inside an iterable (another list) for the chain to work.
        rows = (
            [
                str(obj.pk),
                obj.created.isoformat(),
                str(obj.profile.user.username),
                obj.currency,
                obj.total,
            ]
            for obj in object_list
        )
        return chain(header, rows)



import logging

from django.contrib import admin
from django.db.models import Count

from vendor.models import (
    Address,
    CustomerProfile,
    Invoice,
    Offer,
    OrderItem,
    Payment,
    Price,
    Receipt,
    Subscription,
    TaxClassifier,
    Wishlist,
    WishlistItem,
)
from vendor.models.choice import InvoiceStatus

logger = logging.getLogger(__name__)


################
# ADMIN ACTIONS
################
def soft_delete_payments_with_no_receipt(modeladmin, request, queryset):
    """
    This action finds any successfull payment that has not a receipt that matches its transaction and created fields, and soft deletes them.  # noqa: E501
    Every successfull payment has a receipt that has the same transaction value.
    """
    invalid_payments = [
        payment
        for payment in Payment.objects.filter(success=True)
        if payment.get_receipt() is None
    ]

    for payment in invalid_payments:
        payment.delete()

    logger.info(f"Soft deleted {len(invalid_payments)} payments")


def soft_delete_payments_without_order_items_invoice(modeladmin, request, queryset):
    """
    This action finds any successfull payment that with an invoice that has no order_items.
    Meaning it is a useless payment as it did not pay for any item
    """
    invalid_payments = (
        Payment.objects.filter(success=True)
        .annotate(order_item_count=Count("invoice__order_items"))
        .filter(order_item_count=0)
    )
    for payment in invalid_payments:
        payment.delete()

    logger.info(f"Soft deleted {len(invalid_payments)} payments")


def soft_delete_invoices_with_deleted_payments(modeladmin, request, queryset):
    """
    This action finds any invoice that has a status greater then Checkout and has a soft deleted payment.
    If an invoice has two payments, both need to be deleted to delete the invoice.
    """
    for invoice in Invoice.objects.filter(status__gt=InvoiceStatus.CHECKOUT):
        soft_delete = True
        for payment in invoice.payments.all():
            if not payment.deleted:
                soft_delete = False

        if soft_delete:
            invoice.delete()


soft_delete_payments_with_no_receipt.short_description = (
    "Soft Delete Payments with no receipt"
)

soft_delete_payments_without_order_items_invoice.short_description = (
    "Soft Delete Payments with zero order_items"
)

soft_delete_invoices_with_deleted_payments.short_description = (
    "Soft Delete Invoices with payments that have been soft deleted"
)


###############
# INLINES
###############
class InvoiceInline(admin.TabularInline):
    model = Invoice
    extra = 1


class OrderItemInline(admin.TabularInline):
    model = OrderItem
    extra = 1


class PriceInline(admin.TabularInline):
    model = Price
    extra = 1


class ReceiptInline(admin.TabularInline):
    model = Receipt
    exclude = ("deleted", "uuid", "vendor_notes", "meta", "profile")
    readonly_fields = ("pk", "transaction", "order_item", "start_date", "end_date")
    max_num = 1


class PaymentInline(admin.TabularInline):
    model = Payment
    exclude = (
        "deleted",
        "uuid",
        "provider",
        "profile",
        "billing_address",
        "result",
        "payee_full_name",
        "payee_company",
    )
    readonly_fields = (
        "pk",
        "transaction",
        "invoice",
        "created",
        "status",
        "amount",
        "success",
    )
    max_num = 1


class WishlistInline(admin.TabularInline):
    model = Wishlist
    extra = 1


class WishlistItemInline(admin.TabularInline):
    model = WishlistItem
    extra = 1


###############
# MODEL ADMINS
###############
class AddressAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid",)
    list_display = ("name", "profile", "postal_code")
    search_fields = ("postal_code",)


class CustomerProfileAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid", "user")
    list_display = ("pk", "user", "email", "site", "currency", "created")
    search_fields = ("user__username", "user__email")
    list_filter = ("site__domain",)


class InvoiceAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid", "shipping_address", "profile")
    list_display = (
        "pk",
        "__str__",
        "profile",
        "site",
        "status",
        "total",
        "created",
        "deleted",
    )
    search_fields = (
        "uuid",
        "profile__user__username",
    )
    list_filter = ("site__domain",)
    inlines = [
        OrderItemInline,
    ]
    actions = [soft_delete_invoices_with_deleted_payments]


class OfferAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid",)
    list_display = ("name", "slug", "site", "terms", "available", "deleted")
    search_fields = (
        "name",
        "site__domain",
    )
    list_filter = ("site__domain",)
    inlines = [
        PriceInline,
    ]


class PaymentAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid", "invoice", "profile")
    list_display = (
        "pk",
        "created",
        "submitted_date",
        "transaction",
        "subscription",
        "invoice",
        "profile",
        "amount",
        "deleted",
        "status",
    )
    search_fields = (
        "pk",
        "transaction",
        "profile__user__username",
    )
    list_filter = ("profile__site__domain", "success", "status")
    exclude = ("billing_address",)
    actions = [
        soft_delete_payments_without_order_items_invoice,
        soft_delete_payments_with_no_receipt,
    ]


class ReceiptAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid", "order_item", "profile", "products")
    exclude = ("updated",)
    list_display = (
        "pk",
        "transaction",
        "subscription",
        "created",
        "profile",
        "order_item",
        "start_date",
        "end_date",
        "deleted",
    )
    list_filter = ("profile__site__domain", "products")
    search_fields = (
        "pk",
        "transaction",
        "profile__user__username",
    )
    # Example for future filters on AdminForm Fields
    # def get_form(self, request, obj=None, change=False, **kwargs):
    #     form = super().get_form(request, obj, change, **kwargs)

    #     if not obj:
    #         return form

    #     form.base_fields['profile'].queryset = CustomerProfile.objects.filter(site=obj.profile.site)

    #     return form


class SubscriptionAdmin(admin.ModelAdmin):
    readonly_fields = ("uuid", "profile")
    exclude = ("updated",)
    list_display = ("pk", "gateway_id", "created", "profile", "status")
    list_filter = ("profile__site__domain",)
    search_fields = (
        "pk",
        "gateway_id",
        "profile__user__username",
    )
    inlines = [PaymentInline, ReceiptInline]


class PriceAdmin(admin.ModelAdmin):
    readonly_fields = ("offer",)
    list_display = ("pk", "offer", "cost")
    list_filter = ("offer", "offer__site")
    search_fields = ("pk", "cost", "offer__name")


class TaxClassifierAdmin(admin.ModelAdmin):
    pass


class WishlistAdmin(admin.ModelAdmin):
    inlines = [
        WishlistItemInline,
    ]


###############
# REGISTRATION
###############


admin.site.register(TaxClassifier, TaxClassifierAdmin)
admin.site.register(CustomerProfile, CustomerProfileAdmin)
admin.site.register(Offer, OfferAdmin)
admin.site.register(Invoice, InvoiceAdmin)
admin.site.register(Wishlist, WishlistAdmin)
admin.site.register(Address, AddressAdmin)
admin.site.register(Receipt, ReceiptAdmin)
admin.site.register(Payment, PaymentAdmin)
admin.site.register(Subscription, SubscriptionAdmin)
admin.site.register(OrderItem)
admin.site.register(Price, PriceAdmin)



from django.urls import path

from vendor.views import config as config_views
from vendor.views import integration as integrations_views
from vendor.views import report as report_views
from vendor.views import vendor_admin as admin_views

app_name = "vendor_admin"

urlpatterns = [
    path("", admin_views.AdminDashboardView.as_view(), name="manager-dashboard"),
    path(
        "products/",
        admin_views.AdminProductListView.as_view(),
        name="manager-product-list",
    ),
    path(
        "product/<uuid:uuid>/",
        admin_views.AdminProductUpdateView.as_view(),
        name="manager-product-update",
    ),
    path(
        "product/",
        admin_views.AdminProductCreateView.as_view(),
        name="manager-product-create",
    ),
    path(
        "offers/", admin_views.AdminOfferListView.as_view(), name="manager-offer-list"
    ),
    path(
        "offer/<uuid:uuid>/",
        admin_views.AdminOfferUpdateView.as_view(),
        name="manager-offer-update",
    ),
    path(
        "offer/",
        admin_views.AdminOfferCreateView.as_view(),
        name="manager-offer-create",
    ),
    path(
        "orders/", admin_views.AdminInvoiceListView.as_view(), name="manager-order-list"
    ),
    path(
        "order/<uuid:uuid>/",
        admin_views.AdminInvoiceDetailView.as_view(),
        name="manager-order-detail",
    ),
    path(
        "subscriptions/",
        admin_views.AdminSubscriptionListView.as_view(),
        name="manager-subscriptions",
    ),
    path(
        "stripe/subscriptions/",
        admin_views.AdminStripeSubscriptionListView.as_view(),
        name="manager-stripe-subscriptions",
    ),
    path(
        "stripe/subscription/<uuid:uuid>/recreate/",
        admin_views.AdminStripeSubscriptionReCreate.as_view(),
        name="manager-stripe-subscription-recreate",
    ),
    path(
        "subscription/<uuid:uuid>",
        admin_views.AdminSubscriptionDetailView.as_view(),
        name="manager-subscription",
    ),
    path(
        "subscription/<uuid:uuid_subscription>/add/payment/<uuid:uuid_profile>/",
        admin_views.AdminSubscriptionAddPaymentView.as_view(),
        name="manager-subscription-add-payment",
    ),
    path(
        "subscription/create/",
        admin_views.AdminSubscriptionCreateView.as_view(),
        name="manager-subscription-create",
    ),
    path(
        "profiles/", admin_views.AdminProfileListView.as_view(), name="manager-profiles"
    ),
    path(
        "profile/<uuid:uuid>",
        admin_views.AdminProfileDetailView.as_view(),
        name="manager-profile",
    ),
    path(
        "product/<uuid:uuid>/renew",
        admin_views.AdminManualSubscriptionRenewal.as_view(),
        name="manager-product-renew",
    ),
    path(
        "payments/no/receipt/",
        admin_views.PaymentWithNoReceiptListView.as_view(),
        name="manager-payment-no-receipt",
    ),
    path(
        "payments/no/orderitems/",
        admin_views.PaymentWithNoOrderItemsListView.as_view(),
        name="manager-payment-no-receipt",
    ),
    # Reports
    path(
        "reports/receipts/download/",
        report_views.ReceiptListCSV.as_view(),
        name="manager-receipt-download",
    ),
    path(
        "reports/invoices/download/",
        report_views.InvoiceListCSV.as_view(),
        name="manager-invoice-download",
    ),
    # Integrations
    path(
        "authorizenet/integration/",
        integrations_views.AuthorizeNetIntegrationView.as_view(),
        name="authorizenet-integration",
    ),
    path(
        "stripe/integration/",
        integrations_views.StripeIntegrationView.as_view(),
        name="stripe-integration",
    ),
    # Configs
    path(
        "config/stripe/connect/list/",
        config_views.StripeConnectAccountConfigListView.as_view(),
        name="manager-config-stripe-connect-list",
    ),
    path(
        "config/stripe/connect/create/",
        config_views.StripeConnectAccountCreateConfigView.as_view(),
        name="manager-config-stripe-connect-create",
    ),
    path(
        "config/stripe/connect/update/<int:pk>/",
        config_views.StripeConnectAccountUpdateConfigView.as_view(),
        name="manager-config-stripe-connect-update",
    ),
    path(
        "config/commission/list/",
        config_views.VendorSiteCommissionConfigListView.as_view(),
        name="manager-config-commission-list",
    ),
    path(
        "config/commission/create/",
        config_views.VendorSiteCommissionCreateConfigView.as_view(),
        name="manager-config-commission-create",
    ),
    path(
        "config/commission/update/<int:pk>/",
        config_views.VendorSiteCommissionUpdateConfigView.as_view(),
        name="manager-config-commission-update",
    ),
    # path('config/vendor/commission/', config_views.InvoiceListCSV.as_view(), name="manager-invoice-download"),
    path(
        "config/processor/list/",
        config_views.PaymentProcessorSiteConfigsListView.as_view(),
        name="manager-config-processor-list",
    ),
    path(
        "config/processor/create/",
        config_views.PaymentProcessorCreateConfigView.as_view(),
        name="manager-config-processor-create",
    ),
    path(
        "config/processor/update/<int:pk>",
        config_views.PaymentProcessorUpdateConfigView.as_view(),
        name="manager-config-processor-update",
    ),
]



"""
The encoder and decoder are used for data that gets temporarily stored in the user's session.

To use set the VENDOR_DATA_ENCODER to 'vendor.encrypt.cleartext' in settings.py

This is a 'passthrough encoder' (which is the default) for development.  It should be replaced with a proper encoder to protect user data.  # noqa: E501

To make your own, the module expects to have 'encode' and 'decode' functions.  Everything else is up to you.
"""


def encode(data):
    return data


def decode(data):
    return data



from django.urls import path

from vendor.api.v1 import views as api_views
from vendor.api.v1.authorizenet import views as authorizenet_views
from vendor.api.v1.stripe import views as stripe_views

app_name = "vendor_api"

urlpatterns = [
    path("", api_views.VendorIndexAPI.as_view(), name="api-index"),
    path(
        "cart/add/<slug:slug>/", api_views.AddToCartView.as_view(), name="add-to-cart"
    ),
    path(
        "cart/remove/<slug:slug>/",
        api_views.RemoveFromCartView.as_view(),
        name="remove-from-cart",
    ),
    path(
        "customer/subscription/<uuid:uuid>/cancel/",
        api_views.PaymentGatewaySubscriptionCancelView.as_view(),
        name="customer-subscription-cancel",
    ),
    path(
        "customer/subscription/cancel/model/",
        api_views.CustomerSubscriptionsCancelModel.as_view(),
        name="manager-customer-subscriptions-cancel-model",
    ),
    path(
        "product/<uuid:uuid>/remove",
        api_views.VoidProductView.as_view(),
        name="manager-profile-remove-product",
    ),
    path(
        "profile/<uuid:uuid_profile>/offer/<uuid:uuid_offer>/add",
        api_views.AddOfferToProfileView.as_view(),
        name="manager-profile-add-offer",
    ),
    path(
        "product/<uuid:uuid>/availability",
        api_views.ProductAvailabilityToggleView.as_view(),
        name="manager-product-availablility",
    ),
    path(
        "subscription/price/update",
        api_views.SubscriptionPriceUpdateView.as_view(),
        name="manager-subscription-price-update",
    ),
    path(
        "refund-payment/<uuid:uuid>/",
        api_views.RefundPaymentAPIView.as_view(),
        name="refund-payment-api",
    ),
    # AuthorizeNet
    path(
        "authorizenet/authcapture",
        authorizenet_views.AuthorizeCaptureAPI.as_view(),
        name="api-authorizenet-authcapture",
    ),
    path(
        "authorizenet/void",
        authorizenet_views.VoidAPI.as_view(),
        name="api-authorizenet-void",
    ),
    path(
        "authorizenet/sync/subscriptions/",
        authorizenet_views.SyncSubscriptionsView.as_view(),
        name="api-authorizenet-sync-subscriptions",
    ),
    path(
        "authorizenet/settled/transactions/",
        authorizenet_views.GetSettledTransactionsView.as_view(),
        name="api-authorizenet-settled-transactions",
    ),
    # Stripe
    path(
        "stripe/invoice/paid/",
        stripe_views.StripeInvoicePaid.as_view(),
        name="api-stripe-invoice-paid",
    ),
    path(
        "stripe/invoice/payment/succeded/",
        stripe_views.StripeInvoicePaymentSuccededEvent.as_view(),
        name="api-stripe-invoice-payment-succeded",
    ),
    path(
        "stripe/subscription/invoice/paid/",
        stripe_views.StripeSubscriptionInvoicePaid.as_view(),
        name="api-stripe-subscription-invoice-paid",
    ),
    path(
        "stripe/subscription/invoice/payment/failed/",
        stripe_views.StripeSubscriptionPaymentFailed.as_view(),
        name="api-stripe-subscription-invoice-payment-failed",
    ),
    path(
        "stripe/sync/objects/",
        stripe_views.StripeSyncObjects.as_view(),
        name="api-stripe-sync-objects",
    ),
    path(
        "stripe/card/expiring/",
        stripe_views.StripeCardExpiring.as_view(),
        name="api-stripe-card-expiring",
    ),
    path(
        "stripe/invoice/upcoming/",
        stripe_views.StripeInvoiceUpcomingEvent.as_view(),
        name="api-stripe-invoice-upcoming",
    ),
]



import uuid
from decimal import Decimal

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db import models
from django.db.models.aggregates import Sum
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from vendor.models.base import SoftDeleteModelBase
from vendor.models.choice import PurchaseStatus, RefundReasons
from vendor.models.receipt import Receipt
from vendor.models.subscription import Subscription
from vendor.utils import get_display_decimal


class PaymentReportModelManager(models.Manager):
    def get_total_settled_purchases(self, start_date=None, end_date=None):
        qs = super().get_queryset()

        if not (start_date and end_date):
            return qs.filter(status=PurchaseStatus.SETTLED).aggregate(Sum("amount"))

        if start_date and end_date is None:
            return qs.filter(
                submitted_date__gte=start_date, status=PurchaseStatus.SETTLED
            ).aggregate(Sum("amount"))

        if start_date is None and end_date:
            return qs.filter(
                submitted_date__lte=end_date, status=PurchaseStatus.SETTLED
            ).aggregate(Sum("amount"))

        return qs.filter(
            submitted_date__gte=start_date,
            submitted_date__lte=end_date,
            status=PurchaseStatus.SETTLED,
        ).aggregate((Sum("amount")))

    def get_total_settled_purchases_by_site(self, site, start_date=None, end_date=None):
        qs = super().get_queryset()

        if not (start_date and end_date):
            amount_qs = qs.filter(
                profile__site=site, status=PurchaseStatus.SETTLED
            ).aggregate(Sum("amount"))

        elif start_date and end_date is None:
            amount_qs = qs.filter(
                profile__site=site,
                submitted_date__gte=start_date,
                status=PurchaseStatus.SETTLED,
            ).aggregate(Sum("amount"))

        elif start_date is None and end_date:
            amount_qs = qs.filter(
                profile__site=site,
                submitted_date__lte=end_date,
                status=PurchaseStatus.SETTLED,
            ).aggregate(Sum("amount"))

        else:
            amount_qs = qs.filter(
                profile__site=site,
                submitted_date__gte=start_date,
                submitted_date__lte=end_date,
                status=PurchaseStatus.SETTLED,
            ).aggregate(Sum("amount"))

        if not amount_qs["amount__sum"]:
            return 0

        return amount_qs["amount__sum"]

    def get_total_settled_purchases_by_subscription(
        self, site, start_date=None, end_date=None
    ):
        qs = super().get_queryset()
        organized_data = {}

        if not (start_date and end_date):
            filtered_data = (
                qs.filter(profile__site=site, status=PurchaseStatus.SETTLED)
                .values("subscription")
                .annotate(Sum("amount"))
            )

        elif start_date and end_date is None:
            filtered_data = (
                qs.filter(
                    profile__site=site,
                    submitted_date__gte=start_date,
                    status=PurchaseStatus.SETTLED,
                )
                .values("subscription")
                .annotate(Sum("amount"))
            )

        elif start_date is None and end_date:
            filtered_data = (
                qs.filter(
                    profile__site=site,
                    submitted_date__lte=end_date,
                    status=PurchaseStatus.SETTLED,
                )
                .values("subscription")
                .annotate(Sum("amount"))
            )
        else:
            filtered_data = (
                qs.filter(
                    profile__site=site,
                    submitted_date__gte=start_date,
                    submitted_date__lte=end_date,
                    status=PurchaseStatus.SETTLED,
                )
                .values("subscription")
                .annotate((Sum("amount")))
            )

        organized_data = {}
        for data in filtered_data:
            organized_data[Subscription.objects.get(pk=data["subscription"]).name] = 0

        for data in filtered_data:
            organized_data[
                Subscription.objects.get(pk=data["subscription"]).name
            ] += data["amount__sum"]

        return organized_data

    def get_total_settled_purchases_by_site_and_subscription(
        self, start_date=None, end_date=None
    ):
        qs = super().get_queryset()
        organized_data = {}

        if not (start_date and end_date):
            filtered_data = (
                qs.filter(status=PurchaseStatus.SETTLED)
                .values("profile__site", "subscription")
                .annotate(Sum("amount"))
            )

        elif start_date and end_date is None:
            filtered_data = (
                qs.filter(submitted_date__gte=start_date, status=PurchaseStatus.SETTLED)
                .values("profile__site", "subscription")
                .annotate(Sum("amount"))
            )

        elif start_date is None and end_date:
            filtered_data = (
                qs.filter(submitted_date__lte=end_date, status=PurchaseStatus.SETTLED)
                .values("profile__site", "subscription")
                .annotate(Sum("amount"))
            )

        else:
            filtered_data = (
                qs.filter(
                    submitted_date__gte=start_date,
                    submitted_date__lte=end_date,
                    status=PurchaseStatus.SETTLED,
                )
                .values("profile__site", "subscription")
                .annotate((Sum("amount")))
            )

        organized_data = {data["profile__site"]: {} for data in filtered_data}

        for data in filtered_data:
            organized_data[data["profile__site"]][
                Subscription.objects.get(pk=data["subscription"]).name
            ] = 0

        for data in filtered_data:
            organized_data[data["profile__site"]][
                Subscription.objects.get(pk=data["subscription"]).name
            ] += data["amount__sum"]

        return organized_data


##########
# PAYMENT
##########
class Payment(SoftDeleteModelBase):
    """
    Payments
    - Payments are typically from a Credit Card, PayPal or ACH
    - Multiple Payments can be applied to an invoice
    - Gift cards can be used as payments
    - Discounts are Payment credits
    """

    uuid = models.UUIDField(
        _("UUID"),
        editable=False,
        unique=True,
        default=uuid.uuid4,
        null=False,
        blank=False,
    )
    invoice = models.ForeignKey(
        "vendor.Invoice",
        verbose_name=_("Invoice"),
        on_delete=models.CASCADE,
        related_name="payments",
    )
    created = models.DateTimeField(_("Date Created"), auto_now_add=True)
    submitted_date = models.DateTimeField(
        _("Payment Date"), default=None, blank=True, null=True
    )
    transaction = models.CharField(
        _("Transaction ID"), max_length=80, blank=True, null=True
    )
    provider = models.CharField(
        _("Payment Provider"), max_length=30, blank=True, null=True
    )
    amount = models.FloatField(_("Amount"))
    profile = models.ForeignKey(
        "vendor.CustomerProfile",
        verbose_name=_("Purchase Profile"),
        blank=True,
        on_delete=models.CASCADE,
        related_name="payments",
    )
    billing_address = models.ForeignKey(
        "vendor.Address",
        verbose_name=_("Billing Address"),
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    result = models.JSONField(_("Result"), default=dict, blank=True, null=True)
    success = models.BooleanField(_("Successful"), default=False)
    payee_full_name = models.CharField(_("Name on Card"), max_length=50)
    payee_company = models.CharField(_("Company"), max_length=50, blank=True, null=True)
    status = models.IntegerField(_("Status"), choices=PurchaseStatus.choices, default=0)
    subscription = models.ForeignKey(
        "vendor.Subscription",
        verbose_name=_("Subscription"),
        on_delete=models.CASCADE,
        related_name="payments",
        blank=True,
        null=True,
        default=None,
    )

    reports = PaymentReportModelManager()

    def __str__(self):
        return f"{self.transaction} - {self.profile.user.username}"

    def get_related_receipts(self):
        return Receipt.objects.filter(transaction=self.transaction)

    def get_receipt(self):
        try:
            return Receipt.objects.get(
                transaction=self.transaction,
                created__year=self.created.year,
                created__month=self.created.month,
                created__day=self.created.day,
            )
        except ObjectDoesNotExist:
            return None
        except MultipleObjectsReturned:
            return Receipt.objects.filter(
                transaction=self.transaction,
                created__year=self.created.year,
                created__month=self.created.month,
                created__day=self.created.day,
            ).first()

    def get_amount_display(self):
        return get_display_decimal(self.amount)

    def record_refund(
        self, amount, date=timezone.now(), reason=RefundReasons.OTHER.value
    ):
        self.status = PurchaseStatus.REFUNDED
        if "refunds" not in self.result:
            self.result["refunds"] = []

        self.result["refunds"].append(
            {"date": date.isoformat(), "reason": reason, "amount": str(amount)}
        )

        self.save()

    def is_refund_available(self, refund_amount):
        past_refunds = 0
        past_refund_amount = 0

        if refund_amount > self.amount:
            return False

        if past_refunds := self.result.get("refunds", []):
            past_refund_amount = sum(
                [Decimal(past_refund.get("amount", 0)) for past_refund in past_refunds]
            )

        if (past_refund_amount + refund_amount) > self.amount:
            return False

        return True

    def get_past_refunds(self):
        past_refunds = []
        reasons = {reason[0]: reason[1] for reason in RefundReasons.choices}

        for refund in self.result.get("refunds", []):
            refund["reason"] = reasons[refund["reason"].lower()]
            past_refunds.append(refund)

        return past_refunds



# Generated by Django 3.1.3 on 2020-12-03 19:51
import uuid

from django.db import migrations


def generate_payment_uuid(apps, schema_editor):
    PaymentModel = apps.get_model("vendor", "Payment")
    for payment in PaymentModel.objects.all():
        payment.uuid = uuid.uuid4()
        payment.save(update_fields=["uuid"])


def generate_receipt_uuid(apps, schema_editor):
    ReceiptModel = apps.get_model("vendor", "Receipt")
    for receipt in ReceiptModel.objects.all():
        receipt.uuid = uuid.uuid4()
        receipt.save(update_fields=["uuid"])


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0014_term_types_update_value"),
    ]

    operations = [
        migrations.RunPython(
            generate_payment_uuid, reverse_code=migrations.RunPython.noop
        ),
        migrations.RunPython(
            generate_receipt_uuid, reverse_code=migrations.RunPython.noop
        ),
    ]



# Generated by Django 3.1 on 2020-10-28 17:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0006_auto_20201005_2257"),
    ]

    operations = [
        migrations.AddField(
            model_name="offer",
            name="list_bundle_items",
            field=models.BooleanField(
                default=False,
                help_text="When showing to customers, display the included items in a list?",
                verbose_name="List Bundled Items",
            ),
        ),
    ]



from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic.edit import FormView

from vendor.forms import AuthorizeNetIntegrationForm, StripeIntegrationForm
from vendor.integrations import AuthorizeNetIntegration, StripeIntegration
from vendor.views.mixin import get_site_from_request


class AuthorizeNetIntegrationView(LoginRequiredMixin, FormView):
    template_name = "vendor/integration_form.html"
    form_class = AuthorizeNetIntegrationForm
    success_url = reverse_lazy("vendor_admin:authorizenet-integration")

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        authorizenet_integration = AuthorizeNetIntegration(
            get_site_from_request(self.request)
        )
        context["integration_name"] = _("AuthorizeNet Integration")

        context["form"] = AuthorizeNetIntegrationForm(
            instance=authorizenet_integration.instance
        )

        return context

    def form_valid(self, form):
        authorizenet_integration = AuthorizeNetIntegration(
            get_site_from_request(self.request)
        )
        authorizenet_integration.save(form.cleaned_data)

        return super().form_valid(form)


class StripeIntegrationView(LoginRequiredMixin, FormView):
    template_name = "vendor/integration_form.html"
    form_class = StripeIntegrationForm
    success_url = reverse_lazy("vendor_admin:stripe-integration")

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        stripe_integration = StripeIntegration(get_site_from_request(self.request))
        context["integration_name"] = _("Stripe Integration")

        context["form"] = StripeIntegrationForm(instance=stripe_integration.instance)

        return context

    def form_valid(self, form):
        stripe_integration = StripeIntegration(get_site_from_request(self.request))
        stripe_integration.save(form.cleaned_data)

        return super().form_valid(form)



# Generated by Django 3.1 on 2020-09-15 18:39

import django.db.models.deletion
from django.db import migrations, models

import vendor.models.utils


class Migration(migrations.Migration):

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("vendor", "0002_auto_20200912_0142"),
    ]

    operations = [
        migrations.AlterField(
            model_name="customerprofile",
            name="site",
            field=models.ForeignKey(
                default=vendor.models.utils.set_default_site_id,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="customer_profile",
                to="sites.site",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="site",
            field=models.ForeignKey(
                default=vendor.models.utils.set_default_site_id,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="invoices",
                to="sites.site",
            ),
        ),
        migrations.AlterField(
            model_name="offer",
            name="site",
            field=models.ForeignKey(
                default=vendor.models.utils.set_default_site_id,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="product_offers",
                to="sites.site",
            ),
        ),
    ]



# Generated by Django 3.1.3 on 2021-04-08 23:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0020_lock_uuid_address_profile"),
    ]

    operations = [
        migrations.AlterField(
            model_name="address",
            name="country",
            field=models.IntegerField(
                choices=[
                    (20, "Andorra"),
                    (784, "United Arab Emirates"),
                    (4, "Afghanistan"),
                    (28, "Antigua And Barbuda"),
                    (660, "Anguilla"),
                    (8, "Albania"),
                    (51, "Armenia"),
                    (530, "Netherlands Antilles"),
                    (24, "Angola"),
                    (10, "Antarctica"),
                    (32, "Argentina"),
                    (16, "American Samoa"),
                    (40, "Austria"),
                    (36, "Australia"),
                    (533, "Aruba"),
                    (31, "Azerbaijan"),
                    (70, "Bosnia And Herzegowina"),
                    (52, "Barbados"),
                    (50, "Bangladesh"),
                    (56, "Belgium"),
                    (854, "Burkina Faso"),
                    (100, "Bulgaria"),
                    (48, "Bahrain"),
                    (108, "Burundi"),
                    (204, "Benin"),
                    (60, "Bermuda"),
                    (96, "Brunei Darussalam"),
                    (68, "Bolivia"),
                    (76, "Brazil"),
                    (44, "Bahamas"),
                    (64, "Bhutan"),
                    (74, "Bouvet Island"),
                    (72, "Botswana"),
                    (112, "Belarus"),
                    (84, "Belize"),
                    (124, "Canada"),
                    (166, "Cocos (Keeling) Islands"),
                    (180, "The Democratic Republic Of The Congo"),
                    (140, "Central African Republic"),
                    (178, "Congo"),
                    (756, "Switzerland"),
                    (384, "Cote D'ivoire"),
                    (184, "Cook Islands"),
                    (152, "Chile"),
                    (120, "Cameroon"),
                    (156, "China"),
                    (170, "Colombia"),
                    (188, "Costa Rica"),
                    (192, "Cuba"),
                    (132, "Cape Verde"),
                    (162, "Christmas Island"),
                    (196, "Cyprus"),
                    (203, "Czech Republic"),
                    (276, "Germany"),
                    (262, "Djibouti"),
                    (208, "Denmark"),
                    (212, "Dominica"),
                    (214, "Dominican Republic"),
                    (12, "Algeria"),
                    (218, "Ecuador"),
                    (233, "Estonia"),
                    (818, "Egypt"),
                    (732, "Western Sahara"),
                    (232, "Eritrea"),
                    (724, "Spain"),
                    (231, "Ethiopia"),
                    (246, "Finland"),
                    (242, "Fiji"),
                    (238, "Falkland Islands (Malvinas)"),
                    (583, "Federated States Of Micronesia"),
                    (234, "Faroe Islands"),
                    (250, "France"),
                    (249, "France, Metropolitan"),
                    (266, "Gabon"),
                    (826, "United Kingdom"),
                    (308, "Grenada"),
                    (268, "Georgia"),
                    (254, "French Guiana"),
                    (288, "Ghana"),
                    (292, "Gibraltar"),
                    (304, "Greenland"),
                    (270, "Gambia"),
                    (324, "Guinea"),
                    (312, "Guadeloupe"),
                    (226, "Equatorial Guinea"),
                    (300, "Greece"),
                    (239, "South Georgia And The South Sandwich Islands"),
                    (320, "Guatemala"),
                    (316, "Guam"),
                    (624, "Guinea-bissau"),
                    (328, "Guyana"),
                    (344, "Hong Kong"),
                    (334, "Heard And Mc Donald Islands"),
                    (340, "Honduras"),
                    (191, "Croatia (Local Name: Hrvatska)"),
                    (332, "Haiti"),
                    (348, "Hungary"),
                    (360, "Indonesia"),
                    (372, "Ireland"),
                    (376, "Israel"),
                    (356, "India"),
                    (86, "British Indian Ocean Territory"),
                    (368, "Iraq"),
                    (364, "Iran (Islamic Republic Of Iran)"),
                    (352, "Iceland"),
                    (380, "Italy"),
                    (388, "Jamaica"),
                    (400, "Jordan"),
                    (392, "Japan"),
                    (404, "Kenya"),
                    (417, "Kyrgyzstan"),
                    (116, "Cambodia"),
                    (296, "Kiribati"),
                    (174, "Comoros"),
                    (659, "Saint Kitts And Nevis"),
                    (408, "Democratic People's Republic Of Korea"),
                    (410, "Republic Of Korea"),
                    (414, "Kuwait"),
                    (136, "Cayman Islands"),
                    (398, "Kazakhstan"),
                    (418, "Lao People's Democratic Republic"),
                    (422, "Lebanon"),
                    (662, "Saint Lucia"),
                    (438, "Liechtenstein"),
                    (144, "Sri Lanka"),
                    (430, "Liberia"),
                    (426, "Lesotho"),
                    (440, "Lithuania"),
                    (442, "Luxembourg"),
                    (428, "Latvia"),
                    (434, "Libyan Arab Jamahiriya"),
                    (504, "Morocco"),
                    (492, "Monaco"),
                    (498, "Republic Of Moldova"),
                    (450, "Madagascar"),
                    (584, "Marshall Islands"),
                    (807, "The Former Yugoslav Republic Of Macedonia "),
                    (466, "Mali"),
                    (104, "Myanmar"),
                    (496, "Mongolia"),
                    (446, "Macau"),
                    (580, "Northern Mariana Islands"),
                    (474, "Martinique"),
                    (478, "Mauritania"),
                    (500, "Montserrat"),
                    (470, "Malta"),
                    (480, "Mauritius"),
                    (462, "Maldives"),
                    (454, "Malawi"),
                    (484, "Mexico"),
                    (458, "Malaysia"),
                    (508, "Mozambique"),
                    (516, "Namibia"),
                    (540, "New Caledonia"),
                    (562, "Niger"),
                    (574, "Norfolk Island"),
                    (566, "Nigeria"),
                    (558, "Nicaragua"),
                    (528, "Netherlands"),
                    (578, "Norway"),
                    (524, "Nepal"),
                    (520, "Nauru"),
                    (570, "Niue"),
                    (554, "New Zealand"),
                    (512, "Oman"),
                    (591, "Panama"),
                    (604, "Peru"),
                    (258, "French Polynesia"),
                    (598, "Papua New Guinea"),
                    (608, "Philippines"),
                    (586, "Pakistan"),
                    (616, "Poland"),
                    (666, "St. Pierre And Miquelon"),
                    (612, "Pitcairn"),
                    (630, "Puerto Rico"),
                    (620, "Portugal"),
                    (585, "Palau"),
                    (600, "Paraguay"),
                    (634, "Qatar"),
                    (638, "Reunion"),
                    (642, "Romania"),
                    (643, "Russian Federation"),
                    (646, "Rwanda"),
                    (682, "Saudi Arabia"),
                    (90, "Solomon Islands"),
                    (690, "Seychelles"),
                    (729, "Sudan"),
                    (752, "Sweden"),
                    (702, "Singapore"),
                    (654, "St. Helena"),
                    (705, "Slovenia"),
                    (744, "Svalbard And Jan Mayen Islands"),
                    (703, "Slovakia (Slovak Republic)"),
                    (694, "Sierra Leone"),
                    (674, "San Marino"),
                    (686, "Senegal"),
                    (706, "Somalia"),
                    (740, "Suriname"),
                    (678, "Sao Tome And Principe"),
                    (222, "El Salvador"),
                    (760, "Syrian Arab Republic"),
                    (748, "Swaziland"),
                    (796, "Turks And Caicos Islands"),
                    (148, "Chad"),
                    (260, "French Southern Territories"),
                    (768, "Togo"),
                    (764, "Thailand"),
                    (762, "Tajikistan"),
                    (772, "Tokelau"),
                    (795, "Turkmenistan"),
                    (788, "Tunisia"),
                    (776, "Tonga"),
                    (626, "East Timor"),
                    (792, "Turkey"),
                    (780, "Trinidad And Tobago"),
                    (798, "Tuvalu"),
                    (158, "Taiwan, Province Of China"),
                    (834, "United Republic Of Tanzania"),
                    (804, "Ukraine"),
                    (800, "Uganda"),
                    (581, "United States Minor Outlying Islands"),
                    (840, "United States"),
                    (858, "Uruguay"),
                    (860, "Uzbekistan"),
                    (336, "Holy See (Vatican City State)"),
                    (670, "Saint Vincent And The Grenadines"),
                    (862, "Venezuela"),
                    (92, "Virgin Islands (British)"),
                    (850, "Virgin Islands (U.S.)"),
                    (704, "Viet Nam"),
                    (548, "Vanuatu"),
                    (876, "Wallis And Futuna Islands"),
                    (882, "Samoa"),
                    (887, "Yemen"),
                    (175, "Mayotte"),
                    (710, "South Africa"),
                    (894, "Zambia"),
                    (716, "Zimbabwe"),
                ],
                default=840,
                verbose_name="Country/Region",
            ),
        ),
    ]



import itertools
import math
import uuid

from allauth.account.signals import user_logged_in
from django.contrib.sites.managers import CurrentSiteManager
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.dispatch import receiver
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from vendor.config import DEFAULT_CURRENCY
from vendor.models.utils import set_default_site_id
from vendor.utils import get_site_from_request

from .base import CreateUpdateModelBase, SoftDeleteModelBase
from .choice import CURRENCY_CHOICES, InvoiceStatus, TermType
from .offer import Offer


#####################
# INVOICE
#####################
class Invoice(SoftDeleteModelBase, CreateUpdateModelBase):
    """
    An invoice starts off as a Cart until it is puchased, then it becomes an Invoice.
    """

    uuid = models.UUIDField(_("UUID"), default=uuid.uuid4, editable=False, unique=True)
    profile = models.ForeignKey(
        "vendor.CustomerProfile",
        verbose_name=_("Customer Profile"),
        on_delete=models.CASCADE,
        related_name="invoices",
    )
    site = models.ForeignKey(
        Site,
        verbose_name=_("Site"),
        on_delete=models.CASCADE,
        default=set_default_site_id,
        related_name="invoices",
    )  # For multi-site support
    status = models.IntegerField(
        _("Status"), choices=InvoiceStatus.choices, default=InvoiceStatus.CART
    )
    customer_notes = models.JSONField(
        _("Customer Notes"), default=dict, blank=True, null=True
    )
    vendor_notes = models.JSONField(
        _("Vendor Notes"), default=dict, blank=True, null=True
    )
    ordered_date = models.DateTimeField(
        _("Ordered Date"), blank=True, null=True
    )  # When was the purchase made?
    subtotal = models.FloatField(default=0.0)
    tax = models.FloatField(blank=True, null=True)  # Set on checkout
    shipping = models.FloatField(blank=True, null=True)  # Set on checkout
    total = models.FloatField(blank=True, null=True)  # Set on purchase
    currency = models.CharField(
        _("Currency"), max_length=4, choices=CURRENCY_CHOICES, default=DEFAULT_CURRENCY
    )  # User's default currency
    shipping_address = models.ForeignKey(
        "vendor.Address",
        verbose_name=_("Shipping Address"),
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    global_discount = models.FloatField(
        _("Global Discount"), blank=True, null=True, default=0
    )  # Any value that is set in this field will be subtracted

    objects = models.Manager()
    on_site = CurrentSiteManager()

    class Meta:
        verbose_name = "Invoice"
        verbose_name_plural = "Invoices"
        ordering = [
            "-ordered_date",
            "-updated",
        ]  # TODO: [GK-2518] change to use ordered_date.  Invoice ordered_date needs to be updated on successful purchase by the PaymentProcessor.  # noqa: E501

        permissions = (
            ("can_view_site_purchases", "Can view Site Purchases"),
            ("can_refund_purchase", "Can refund Purchase"),
        )

    def __str__(self):
        if not self.profile.user:  # Can this ever even happen?
            return "New Invoice"
        return f"{self.profile.user.username} - {self.uuid}"

    def get_invoice_display(self):
        if self.profile.user.username is None:
            return _(f"Invoice ({self.created:%Y-%m-%d %H:%M})")
        return _(
            f"{self.profile.user.username} Invoice ({self.created:%Y-%m-%d %H:%M})"
        )

    def add_offer(self, offer, quantity=1):
        self.global_discount = 0

        order_item, created = self.order_items.get_or_create(offer=offer)

        # make sure the invoice pk is also in the OriderItem
        if not created and order_item.offer.allow_multiple:
            order_item.quantity += quantity
            order_item.save()

        self.update_totals()
        self.save()

        return order_item

    def remove_offer(self, offer, clear=False):
        self.global_discount = 0
        try:
            order_item = self.order_items.get(
                offer=offer
            )  # Get the order item if it's present
        except ObjectDoesNotExist:
            return 0

        order_item.quantity -= 1

        if order_item.quantity == 0 or clear:
            order_item.delete()
        else:
            order_item.save()

        if not (
            self.get_recurring_order_items().count()
            or self.get_one_time_transaction_order_items().count()
        ):
            for order_item in self.order_items.all():
                order_item.delete()

        self.update_totals()
        self.save()

        return order_item

    def swap_offer(self, existing_offer, new_offer):
        """
        Functions swaps offers that have the same linked product. It will not remove bundle offers
        that also have shared product with the new offer. The function comes in handy to swap
        an offer that has the normal price with one that has a discount price or terms.
        """
        if not existing_offer.products.filter(
            pk__in=[offer.pk for offer in new_offer.products.all()]
        ).exists():
            return None

        if self.order_items.filter(offer=existing_offer).exists():
            order_items_same_product = self.order_items.filter(
                offer=existing_offer
            ).exclude(offer__bundle=True)
            for order_item in order_items_same_product:
                self.remove_offer(order_item.offer, clear=True)

        self.add_offer(new_offer)
        self.update_totals()
        self.save()

    def calculate_shipping(self):
        """
        Based on the Shipping Address
        """
        self.shipping = 0

    def calculate_tax(self):
        """
        Extendable
        """
        self.tax = 0

    def calculate_subtotal(self):
        """
        Get the total amount of the offer, which could be a set price or the products MSRP
        """
        return sum(
            [
                item.total
                for item in self.order_items.exclude(offer__is_promotional=True)
            ]
        )

    def update_totals(self):
        """
        Sets the invoice total field by calculating its subtotal, any discounts, its shipping and tax.
        If by any reason the total is a negative value it will return 0 as vendor cannot credit any acount
        """
        self.subtotal = self.calculate_subtotal()
        discounts = self.get_discounts()
        self.calculate_shipping()
        self.calculate_tax()
        self.total = (self.subtotal - discounts) + self.tax + self.shipping

        if self.total < 0:
            self.total = 0

    def get_payment_billing_address(self):
        if not self.payments.first().billing_address:
            return ""
        return self.payments.first().billing_address.get_address_display()

    def get_absolute_management_url(self):
        """
        This is the url to the management's detail page for the Invoice
        """
        return reverse("vendor_admin:manager-order-detail", kwargs={"uuid": self.uuid})

    def get_recurring_order_items(self):
        """
        Gets the recurring order items in the invoice
        """
        return self.order_items.filter(
            offer__terms__lt=TermType.PERPETUAL, offer__is_promotional=False
        )

    def get_recurring_total(self):
        """
        Gets the total price for all recurring order items in the invoice and subtracting any discounts.
        """
        recurring_time_order_items = self.get_recurring_order_items()
        return sum(
            [
                (order_item.total - order_item.discounts)
                for order_item in recurring_time_order_items.all()
            ]
        )

    def get_one_time_transaction_order_items(self):
        """
        Gets one time transation order items in the invoice
        """
        return self.order_items.filter(
            offer__terms__gte=TermType.PERPETUAL, offer__is_promotional=False
        )

    def get_one_time_transaction_total(self):
        """
        Gets the total price for order items that will be purchased on a single transation. It also subtracts any discounts  # noqa: E501
        """
        one_time_order_items = self.get_one_time_transaction_order_items()
        return sum(
            [
                (order_item.total - order_item.discounts)
                for order_item in one_time_order_items.all()
            ]
        )

    def empty_cart(self):
        """
        Remove any offer/order_item if the invoice is in Cart State.
        """
        offers = []
        offers = list(
            itertools.chain.from_iterable(
                [
                    [order_item.offer] * order_item.quantity
                    for order_item in self.order_items.all()
                ]
            )
        )
        for offer in offers:
            self.remove_offer(offer)

    def get_next_billing_date(self):
        """
        Return the next billing date, if an invoice has two different billing dates it will return
        the upcoming one.
        """
        recurring_offers = self.order_items.filter(
            offer__terms__lt=TermType.PERPETUAL, offer__is_promotional=False
        )

        if not recurring_offers.count():
            return None

        next_billing_dates = [
            order_item.offer.get_next_billing_date() for order_item in recurring_offers
        ]

        next_billing_dates.sort()

        return next_billing_dates[0]

    def get_coupon_code_order_item(self):
        return (
            self.order_items.filter(offer__is_promotional=True).first()
            if self.order_items.filter(offer__is_promotional=True).exists()
            else None
        )

    def get_billing_dates_and_prices(self):
        now = timezone.now()
        payment_dates = {now: self.get_one_time_transaction_total()}

        coupon_order_item = self.get_coupon_code_order_item()

        for recurring_order_item in self.get_recurring_order_items():
            offer_total = recurring_order_item.total

            if coupon_order_item:
                if next(
                    (
                        product
                        for product in coupon_order_item.offer.products.all()
                        if product in recurring_order_item.offer.products.all()
                    ),
                    False,
                ):
                    if coupon_order_item.offer.promo_campaign.first().is_percent_off:
                        offer_total = offer_total - (
                            (offer_total * coupon_order_item.offer.current_price())
                            / 100
                        )
                        # offer_total = offer_total - (recurring_order_item.discounts + ((offer_total * coupon_order_item.offer.current_price()) / 100))  # noqa: E501
                    else:
                        offer_total = offer_total - math.fabs(
                            coupon_order_item.offer.current_price()
                        )

            # if recurring_order_item.discounts or self.global_discount:
            #     offer_total = offer_total - (recurring_order_item.discounts + math.fabs(self.global_discount))

            start_date = recurring_order_item.offer.get_offer_start_date(now)

            if (
                recurring_order_item.offer.has_trial()
                or recurring_order_item.offer.billing_start_date
            ) and not self.profile.has_owned_product(
                recurring_order_item.offer.products.all()
            ):
                start_date = (
                    recurring_order_item.offer.get_payment_start_date_trial_offset(now)
                )

                if recurring_order_item.offer.get_trial_occurrences() > 1:
                    offer_total = recurring_order_item.offer.get_trial_amount()

            if start_date in payment_dates:
                payment_dates.update(
                    {start_date: payment_dates[start_date] + offer_total}
                )
            else:
                payment_dates[start_date] = offer_total

        sorted_payments = {
            key: payment_dates[key] for key in sorted(payment_dates.keys())
        }

        return sorted_payments

    def get_next_billing_price(self):
        """
        Returns the price corresponding to the upcoming billing date.
        """
        recurring_offers = self.order_items.filter(
            offer__terms__lt=TermType.PERPETUAL, offer__is_promotional=False
        )

        if not recurring_offers.count():
            return None

        if recurring_offers.count() == 1:
            return recurring_offers.first().total

        next_billing_date = recurring_offers.first().offer.get_next_billing_date()
        next_billing_date_price = recurring_offers.first().total

        for order_item in recurring_offers:
            if order_item.offer.get_next_billing_date() < next_billing_date:
                next_billing_date = order_item.offer.get_next_billing_date()
                next_billing_date_price = order_item.total

        return next_billing_date_price

    def get_savings(self):
        savings = 0

        savings = self.calculate_subtotal() - self.get_discounts()

        return savings

    def get_discounts(self):
        """
        Returns the sum of discounts and trial_discounts. Discounts are related to the offer.price and the offer.product.msrp,  # noqa: E501
        while trial discounts are related to the set offer.meta.trial_amount if it has a trial_occurrence.
        """
        if "discounts" in self.vendor_notes:
            return self.vendor_notes["discounts"]

        discounts = 0

        discounts = sum(
            [
                order_item.discounts
                for order_item in self.order_items.all()
                if not self.profile.has_owned_product(order_item.offer.products.all())
            ]
        )

        trial_discounts = sum(
            [
                order_item.trial_amount - order_item.price
                for order_item in self.order_items.all()
                if not self.profile.has_owned_product(order_item.offer.products.all())
                and (
                    order_item.offer.has_trial_occurrences()
                    or order_item.offer.get_trial_days()
                )
            ]
        )

        return discounts + math.fabs(trial_discounts) + math.fabs(self.global_discount)

    def get_discounts_display(self):
        """
        Only return promotional, and global discounts. If their is a trial period price it is not added to the display.
        """
        if "discounts" in self.vendor_notes:
            return self.vendor_notes["discounts"]

        discounts = 0
        coupon_code_order_item = self.get_coupon_code_order_item()  # noqa: F841

        discounts = sum(
            [
                order_item.discounts
                for order_item in self.order_items.all()
                if not self.profile.has_owned_product(order_item.offer.products.all())
            ]
        )

        # if coupon_code_order_item and coupon_code_order_item.offer.promo_campaign.first().is_percent_off:
        #     discounts += coupon_code_order_item.offer.current_price()

        return discounts + math.fabs(self.global_discount)

    def save_discounts_vendor_notes(self):
        """
        Once an invoice has been completed, this method saves the discounts applied to the invoice
        in the vendor_notes field. This makes it for faster lookup on get_discounts for future invoice
        views.
        """
        if not isinstance(self.vendor_notes, dict):
            self.vendor_notes = dict()

        self.vendor_notes["discounts"] = self.get_discounts()
        self.save()

    def save_promo_codes(self, codes):
        # TODO: Need to implement a consistant way on how to save promo codes in invoice.vendor_notes
        pass

    def get_promos(self):
        if "promos" not in self.vendor_notes or not len(
            self.vendor_notes.get("promos", [])
        ):
            return ""
        return self.vendor_notes["promos"].keys()

    def clear_promos(self):
        for order_item in self.order_items.filter(offer__is_promotional=True):
            self.remove_offer(order_item.offer, clear=True)

    def get_products(self):
        invoice_products = set(
            [
                product
                for order_item in self.order_items.all()
                for product in order_item.offer.products.all()
            ]
        )

        return list(invoice_products)


class OrderItem(CreateUpdateModelBase):
    """
    A link for each item to a user after it's been purchased
    """

    invoice = models.ForeignKey(
        "vendor.Invoice",
        verbose_name=_("Invoice"),
        on_delete=models.CASCADE,
        related_name="order_items",
    )
    offer = models.ForeignKey(
        "vendor.Offer",
        verbose_name=_("Offer"),
        on_delete=models.CASCADE,
        related_name="order_items",
    )
    quantity = models.IntegerField(_("Quantity"), default=1)

    class Meta:
        verbose_name = "Order Item"
        verbose_name_plural = "Order Items"

    def __str__(self):
        return f"{self.offer} - {self.invoice.uuid}"

    @property
    def total(self):
        return self.quantity * self.price

    @property
    def price(self):
        """
        Price property is calculated if a offer.product has an MSRP different from zero.
        if product MSRP is zero it will return the corresponding offer.price.cost
        """
        if self.offer.get_msrp():
            return self.offer.get_msrp()
        return self.offer.current_price()

    @property
    def name(self):
        return self.offer.name

    @property
    def discounts(self):
        return self.offer.discount() * self.quantity

    @property
    def trial_amount(self):
        if self.receipts.count():
            if "first" in self.receipts.first().meta:
                return self.offer.get_trial_amount()
        else:
            if self.offer.has_trial_occurrences():
                return self.offer.get_trial_amount()
            elif self.offer.get_trial_days():
                return self.offer.get_trial_amount()
        return self.offer.current_price()

    def get_total_display(self):
        if not self.total:
            return "0.00"

        return f"{self.total:2}"


##########
# Signals
##########
@receiver(user_logged_in)
def convert_session_cart_to_invoice(sender, request, **kwargs):
    if "session_cart" in request.session:
        profile, created = request.user.customer_profile.get_or_create(
            site=get_site_from_request(request)
        )
        cart = profile.get_cart()

        for offer_key in request.session["session_cart"].keys():
            cart.add_offer(
                Offer.objects.get(pk=offer_key),
                quantity=request.session["session_cart"][offer_key]["quantity"],
            )

        del request.session["session_cart"]



# Generated by Django 3.1.3 on 2020-12-02 17:59

# import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("vendor", "0012_auto_20201123_1848"),
    ]

    operations = [
        migrations.AddField(
            model_name="payment",
            name="uuid",
            field=models.UUIDField(
                editable=True, unique=False, verbose_name="UUID", null=True
            ),
        ),
        migrations.AddField(
            model_name="receipt",
            name="uuid",
            field=models.UUIDField(
                editable=True, unique=False, verbose_name="UUID", null=True
            ),
        ),
        migrations.AlterField(
            model_name="offer",
            name="terms",
            field=models.IntegerField(
                choices=[
                    (100, "Subscription"),
                    (101, "Monthly Subscription"),
                    (103, "Quarterly Subscription"),
                    (106, "Semi-Annual Subscription"),
                    (112, "Annual Subscription"),
                    (200, "Perpetual"),
                    (220, "One-Time Use"),
                ],
                default=0,
                verbose_name="Terms",
            ),
        ),
    ]
